(* file: filterFold.ml
   author: Bob Muller

   CS1103 Computer Science I Honors

   Lecture Notes for Meeting 2 of Week 6

   Topics:
     1. Filtering
     2. Folding
     3. Application: Rain Tank

  1. Filtering  **********************************************************

  Last time we saw the map function:

  map : ('a -> 'b) -> 'a list -> 'b list
 *)
let rec map f xs =
  match xs with
  | [] -> []
  | x :: xs -> f x :: map f xs

(* Today we'll see two related functions. Let

  isEven : int -> bool
*)
let isEven n = n mod 2 = 0

(* Write a function

   onlyEvens : int list -> int list

   such that a call (onlyEvens ns) returns only the even numbers in ns.
*)
let rec onlyEvens ns =
  match ns with
  | [] -> []
  | n :: ns -> (match isEven n with
                | true  -> n :: onlyEvens ns
                | false -> onlyEvens ns)

(* Let
 *)
type point = {x : int; y : int }

(* isDiagonalPoint : point -> bool
*)
let isDiagonalPoint {x; y} = x = y

(* Write a function

   diagonalPoints : point list -> point list

   such that a given call (diagonalPoints points) returns only those
   points with equal x and y components.
 *)
let rec diagonalPoints points =
  match points with
  | [] -> []
  | point :: points -> (match isDiagonalPoint point with
                        | true  -> point :: diagonalPoints points
                        | false -> diagonalPoints points)

(* As with the map examples, these two pieces of code are identical
   except for the testing function. We abstract by replacing the
  varying parts with a variable:

   filter : ('a -> bool) -> 'a list -> 'a list
 *)
let rec filter test xs =
  match xs with
  | [] -> []
  | x :: xs -> (match test x with
                | true  -> x :: filter test xs
                | false -> filter test xs)

(* Then
*)
let onlyEvens ns = filter isEven ns
let diagonalPoints points = filter isDiagonalPoint points

(* Let

   fromTo : int -> int -> int list
*)
let rec fromTo lo hi =
  match lo > hi with
  | true  -> []
  | false -> lo :: fromTo (lo + 1) hi

(* eratosthenes : int -> int list

   The call (eratosthenes n) returns all of the prime numbers less
   than or equal to n.
*)
let eratosthenes n =
  let rec repeat ns primes =
    match ns with
    | [] -> primes
    | m :: ms ->
       let isntMultiple n = (n mod m) <> 0
       in
       repeat (filter isntMultiple ms) (m :: primes)
  in
  repeat (fromTo 2 n) []

(* 2. Folding    **********************************************************

   Like mapping and filtering, it's common to combine elements of a list
  into single value. Consider:

   sum : int list -> int
*)
let rec sum ns =
  match ns with
  | [] -> 0
  | n :: ns -> (+) n (sum ns)

(* appendAll : ('a list) list -> 'a list
*)
let rec appendAll xxs =
  match xxs with
  | [] -> []
  | xs :: xxs -> (@) xs (appendAll xxs)

(* These two functions vary in two places -- 0 vs [] and (+) vs (@).
  We abstract

   foldRight : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
*)
let rec foldRight op xs id =
  match xs with
  | [] -> id
  | x :: xs -> op x (foldRight op xs id)

(* NB: (foldRight f [a1; ...; an] id) ->> f a1 (f a2 (... (f an id) ...))

  average : int list -> int
*)
let average ns = (foldRight (+) ns 0) / (List.length ns)

(* foldLeft :

  foldLeft : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a

   (foldLeft f id [b1; ...; bn]) ->> f (... (f (f id b1) b2) ...) bn
*)
let rec foldLeft op id xs =
  match xs with
  | [] -> id
  | x :: xs -> foldLeft op xs (op id x)


(* 3. Application: Rain Tank **********************************************

   See the code in the rain/ folder.
*)
